import os
import glob
from qtpy import QtWidgets, QtCore
import numpy as np
import matplotlib.patches as patches
from ..utils import get_temp_dir, extract_extension, InformationBox
from .mplcontroller import MplController
from .ccdprocesscontroller import CCDProcessController
from PIL import Image
import json
import datetime

class CCDController(object):

    def __init__(self, model, widget):
        self.model = model
        self.widget = widget
        self.ccdprocess_ctrl = CCDProcessController(self.model, self.widget)
        self.plot_ctrl = MplController(self.model, self.widget)
        self._row_roi_selector = None
        self._row_roi_press_cid = None
        self._row_roi_motion_cid = None
        self._row_roi_release_cid = None
        self._row_roi_press_y = None
        self._row_roi_preview = None
        self.connect_channel()

    def connect_channel(self):
        if hasattr(self.widget, "pushButton_Info"):
            self.widget.pushButton_Info.clicked.connect(self.show_tif_header)
        if hasattr(self.widget, "pushButton_S_CCDReset"):
            self.widget.pushButton_S_CCDReset.clicked.connect(
                self.reset_max_cake_scale)
        self.widget.pushButton_ApplyCakeView.clicked.connect(self.update_cake)
        self.widget.pushButton_ApplyMask.clicked.connect(self.apply_mask)
        self.widget.pushButton_MaskReset.clicked.connect(self.reset_maskrange)
        self.widget.pushButton_ResetCakeScale.clicked.connect(
            self.reset_max_cake_scale)
        if hasattr(self.widget, "spinBox_CCDRowMin"):
            self.widget.spinBox_CCDRowMin.valueChanged.connect(
                self._on_row_roi_spin_changed)
        if hasattr(self.widget, "spinBox_CCDRowMax"):
            self.widget.spinBox_CCDRowMax.valueChanged.connect(
                self._on_row_roi_spin_changed)
        if hasattr(self.widget, "pushButton_CCDSelectRoi"):
            self.widget.pushButton_CCDSelectRoi.toggled.connect(
                self._toggle_row_roi_selector)
        if hasattr(self.widget, "pushButton_CCDFullRoi"):
            self.widget.pushButton_CCDFullRoi.clicked.connect(
                self._reset_row_roi_to_full)
        if hasattr(self.widget, "comboBox_CakeColormap"):
            self.widget.comboBox_CakeColormap.currentIndexChanged.connect(
                self._apply_changes_to_graph)
        if hasattr(self.widget, "doubleSpinBox_CCDScaleMin"):
            self.widget.doubleSpinBox_CCDScaleMin.valueChanged.connect(
                self._on_cake_scale_spin_changed)
        if hasattr(self.widget, "doubleSpinBox_CCDScaleMax"):
            self.widget.doubleSpinBox_CCDScaleMax.valueChanged.connect(
                self._on_cake_scale_spin_changed)
        if hasattr(self.widget, "cake_hist_widget"):
            self.widget.cake_hist_widget.boundChanged.connect(
                self._set_cake_bound_from_hist)
        """
        self.widget.pushButton_Load_CakeFormatFile.clicked.connect(
            self.load_cake_format_file)
        self.widget.pushButton_Save_CakeFormatFile.clicked.connect(
            self.save_cake_format_file)
        """

    def update_cake(self):
        if not self.model.base_ptn_exist():
            QtWidgets.QMessageBox.warning(
                self.widget, 'Warning', 'Choose a spectrum file first.')
            return
        if (not self._is_spe_source()) and (not self.model.associated_image_exists()):
            self._set_image_file_box_missing()
            QtWidgets.QMessageBox.warning(
                self.widget, 'Warning',
                'CCD image file does not exist in the spectrum folder.\n'
                'Move the image file into the same folder as the spectrum first.')
            return

        success = self.produce_cake()
        if not success:
            return
        temp_dir = get_temp_dir(self.model.get_base_ptn_filename())
        self.model.diff_img.write_temp_cakefiles(temp_dir=temp_dir)
        self._set_image_file_box()
        self._apply_changes_to_graph()

    def _set_row_roi_spin_limits(self):
        if (not hasattr(self.widget, "spinBox_CCDRowMin")) or \
                (not self.model.diff_img_exist()):
            return
        img = getattr(self.model.diff_img, "img", None)
        if img is None or np.ndim(img) < 2 or img.shape[0] <= 0:
            return
        if img.shape[0] <= 1:
            return
        self._ensure_row_roi_defined_for_full_ccd(img)
        n_rows = int(img.shape[0])
        for box in (self.widget.spinBox_CCDRowMin, self.widget.spinBox_CCDRowMax):
            box.blockSignals(True)
            box.setMaximum(n_rows - 1)
            box.blockSignals(False)
        current = getattr(self.model.base_ptn, "row_roi", None)
        if current is None:
            current = (0, n_rows - 1)
        row_min = max(0, min(int(current[0]), n_rows - 1))
        row_max = max(0, min(int(current[1]), n_rows - 1))
        if row_max < row_min:
            row_min, row_max = row_max, row_min
        self.widget.spinBox_CCDRowMin.blockSignals(True)
        self.widget.spinBox_CCDRowMax.blockSignals(True)
        self.widget.spinBox_CCDRowMin.setValue(row_min)
        self.widget.spinBox_CCDRowMax.setValue(row_max)
        self.widget.spinBox_CCDRowMin.blockSignals(False)
        self.widget.spinBox_CCDRowMax.blockSignals(False)

    def _ensure_row_roi_defined_for_full_ccd(self, img=None):
        if not self.model.base_ptn_exist():
            return False
        if img is None:
            img = getattr(getattr(self.model, "diff_img", None), "img", None)
        if img is None or np.ndim(img) < 2 or img.shape[0] <= 0:
            return False
        if img.shape[0] <= 1:
            return False
        current = getattr(self.model.base_ptn, "row_roi", None)
        is_undefined = current is None
        if current is not None:
            try:
                row_min = int(current[0])
                row_max = int(current[1])
                is_undefined = ((row_min, row_max) == (0, 0)) or (row_min == row_max)
            except Exception:
                is_undefined = True
        if not is_undefined:
            return False
        return bool(self.model.base_ptn.set_spe_row_roi(0, int(img.shape[0] - 1)))

    def _apply_row_roi_to_spectrum(self):
        if (not self.model.base_ptn_exist()) or \
                (extract_extension(str(self.model.get_base_ptn_filename())).lower() != 'spe'):
            return
        if (not hasattr(self.widget, "spinBox_CCDRowMin")) or \
                (getattr(self.model.base_ptn, "raw_image", None) is None):
            return
        raw_image = getattr(self.model.base_ptn, "raw_image", None)
        if np.ndim(raw_image) < 2 or raw_image.shape[0] <= 1:
            return
        row_min = int(self.widget.spinBox_CCDRowMin.value())
        row_max = int(self.widget.spinBox_CCDRowMax.value())
        if row_max < row_min:
            row_min, row_max = row_max, row_min
            self.widget.spinBox_CCDRowMin.blockSignals(True)
            self.widget.spinBox_CCDRowMax.blockSignals(True)
            self.widget.spinBox_CCDRowMin.setValue(row_min)
            self.widget.spinBox_CCDRowMax.setValue(row_max)
            self.widget.spinBox_CCDRowMin.blockSignals(False)
            self.widget.spinBox_CCDRowMax.blockSignals(False)
        if row_min == row_max:
            row_min = 0
            row_max = int(raw_image.shape[0] - 1)
            self.widget.spinBox_CCDRowMin.blockSignals(True)
            self.widget.spinBox_CCDRowMax.blockSignals(True)
            self.widget.spinBox_CCDRowMin.setValue(row_min)
            self.widget.spinBox_CCDRowMax.setValue(row_max)
            self.widget.spinBox_CCDRowMin.blockSignals(False)
            self.widget.spinBox_CCDRowMax.blockSignals(False)
        success = self.model.base_ptn.set_spe_row_roi(row_min, row_max)
        if not success:
            return
        self._refresh_bgsub_for_current_spectrum()
        self._apply_changes_to_graph()

    def _refresh_bgsub_for_current_spectrum(self):
        if not self.model.base_ptn_exist():
            return
        x_raw, __ = self.model.base_ptn.get_raw()
        if x_raw is None or len(x_raw) == 0:
            return
        roi_min = float(self.widget.doubleSpinBox_Background_ROI_min.value())
        roi_max = float(self.widget.doubleSpinBox_Background_ROI_max.value())
        if (x_raw.min() >= roi_min) or (x_raw.max() <= roi_min):
            roi_min = float(x_raw.min())
            self.widget.doubleSpinBox_Background_ROI_min.setValue(roi_min)
        if (x_raw.max() <= roi_max) or (x_raw.min() >= roi_max):
            roi_max = float(x_raw.max())
            self.widget.doubleSpinBox_Background_ROI_max.setValue(roi_max)
        params = [
            int(self.widget.spinBox_BGParam1.value()),
        ]
        x_raw, y_raw = self.model.base_ptn.get_raw()
        __, y_fit = self.plot_ctrl._get_smoothed_pattern_xy(x_raw, y_raw)
        if not self.plot_ctrl._smoothing_active():
            y_fit = y_raw
        fit_areas = []
        table = getattr(self.widget, "tableWidget_BackgroundConstraints", None)
        if table is not None:
            for row in range(table.rowCount()):
                item_min = table.item(row, 0)
                item_max = table.item(row, 1)
                if item_min is None or item_max is None:
                    continue
                try:
                    xmin = float(item_min.text())
                    xmax = float(item_max.text())
                except Exception:
                    continue
                if xmax < xmin:
                    xmin, xmax = xmax, xmin
                fit_areas.append([xmin, xmax])
        self.model.base_ptn.get_chbg(
            [roi_min, roi_max], params, yshift=0,
            fit_areas=fit_areas, y_source=y_fit)

    def _on_row_roi_spin_changed(self, value):
        del value
        self._apply_row_roi_to_spectrum()

    def _toggle_row_roi_selector(self, checked):
        if checked:
            self._activate_row_roi_selector()
        else:
            self._deactivate_row_roi_selector()

    def _activate_row_roi_selector(self):
        if not self.model.diff_img_exist():
            self.widget.pushButton_CCDSelectRoi.setChecked(False)
            return
        if not hasattr(self.widget.mpl.canvas, "ax_cake"):
            self.widget.pushButton_CCDSelectRoi.setChecked(False)
            return
        self._deactivate_row_roi_selector()
        toolbar = getattr(self.widget.mpl.canvas, "toolbar", None)
        if toolbar is not None:
            try:
                if getattr(toolbar, "mode", "") == 'zoom rect':
                    toolbar.zoom()
                elif getattr(toolbar, "mode", "") == 'pan/zoom':
                    toolbar.pan()
            except Exception:
                pass
        canvas = self.widget.mpl.canvas
        self._row_roi_press_y = None
        try:
            canvas.setCursor(QtWidgets.QCursor(QtCore.Qt.CrossCursor))
        except Exception:
            pass
        self._row_roi_press_cid = canvas.mpl_connect(
            "button_press_event", self._on_row_roi_press)
        self._row_roi_motion_cid = canvas.mpl_connect(
            "motion_notify_event", self._on_row_roi_motion)
        self._row_roi_release_cid = canvas.mpl_connect(
            "button_release_event", self._on_row_roi_release)
        if hasattr(self.widget, "pushButton_CCDSelectRoi"):
            self.widget.pushButton_CCDSelectRoi.blockSignals(True)
            self.widget.pushButton_CCDSelectRoi.setChecked(True)
            self.widget.pushButton_CCDSelectRoi.blockSignals(False)
        try:
            self.widget.mpl.canvas.draw_idle()
        except Exception:
            pass

    def _deactivate_row_roi_selector(self):
        canvas = getattr(self.widget, "mpl", None)
        canvas = getattr(canvas, "canvas", None)
        if canvas is not None:
            try:
                if self._row_roi_press_cid is not None:
                    canvas.mpl_disconnect(self._row_roi_press_cid)
            except Exception:
                pass
            try:
                if self._row_roi_motion_cid is not None:
                    canvas.mpl_disconnect(self._row_roi_motion_cid)
            except Exception:
                pass
            try:
                if self._row_roi_release_cid is not None:
                    canvas.mpl_disconnect(self._row_roi_release_cid)
            except Exception:
                pass
        if self._row_roi_preview is not None:
            try:
                self._row_roi_preview.remove()
            except Exception:
                pass
        self._row_roi_selector = None
        self._row_roi_press_cid = None
        self._row_roi_motion_cid = None
        self._row_roi_release_cid = None
        self._row_roi_press_y = None
        self._row_roi_preview = None
        canvas = getattr(self.widget, "mpl", None)
        canvas = getattr(canvas, "canvas", None)
        if canvas is not None:
            try:
                canvas.unsetCursor()
            except Exception:
                pass
        if hasattr(self.widget, "pushButton_CCDSelectRoi"):
            self.widget.pushButton_CCDSelectRoi.blockSignals(True)
            self.widget.pushButton_CCDSelectRoi.setChecked(False)
            self.widget.pushButton_CCDSelectRoi.blockSignals(False)

    def _on_row_roi_press(self, event):
        ax = getattr(self.widget.mpl.canvas, "ax_cake", None)
        if event.inaxes != ax:
            return
        if event.button != 1 or event.ydata is None:
            return
        self._row_roi_press_y = float(event.ydata)
        self._update_row_roi_preview(float(event.ydata))

    def _on_row_roi_motion(self, event):
        if self._row_roi_press_y is None:
            return
        ax = getattr(self.widget.mpl.canvas, "ax_cake", None)
        y_now = self._event_y_to_cake_row(event, ax)
        if y_now is None:
            return
        self._set_row_roi_spin_values_from_pixels(self._row_roi_press_y, y_now)
        self._update_row_roi_preview(float(y_now))

    def _on_row_roi_release(self, event):
        if self._row_roi_press_y is None:
            return
        ax = getattr(self.widget.mpl.canvas, "ax_cake", None)
        if event.button != 1:
            self._deactivate_row_roi_selector()
            return
        y_release = self._event_y_to_cake_row(event, ax)
        if y_release is None:
            self._deactivate_row_roi_selector()
            return
        img = getattr(self.model.diff_img, "img", None)
        if img is None or np.ndim(img) < 2:
            self._deactivate_row_roi_selector()
            return
        self._set_row_roi_spin_values_from_pixels(self._row_roi_press_y, y_release)
        self._apply_row_roi_to_spectrum()
        self._deactivate_row_roi_selector()

    def _event_y_to_cake_row(self, event, ax):
        if ax is None:
            return None
        if event.ydata is not None and event.inaxes == ax:
            return float(event.ydata)
        if event.y is None:
            return None
        try:
            __, ydata = ax.transData.inverted().transform((event.x, event.y))
            return float(ydata)
        except Exception:
            return None

    def _update_row_roi_preview(self, y_current):
        ax = getattr(self.widget.mpl.canvas, "ax_cake", None)
        if ax is None:
            return
        try:
            x0, x1 = ax.get_xlim()
        except Exception:
            return
        y0 = min(float(self._row_roi_press_y), float(y_current))
        y1 = max(float(self._row_roi_press_y), float(y_current))
        if self._row_roi_preview is None:
            self._row_roi_preview = patches.Rectangle(
                (x0, y0),
                (x1 - x0),
                max(y1 - y0, 1.0),
                linewidth=1.5,
                edgecolor="#00e676",
                facecolor="#00e676",
                alpha=0.18,
                linestyle="--",
            )
            ax.add_patch(self._row_roi_preview)
        else:
            self._row_roi_preview.set_x(x0)
            self._row_roi_preview.set_y(y0)
            self._row_roi_preview.set_width(x1 - x0)
            self._row_roi_preview.set_height(max(y1 - y0, 1.0))
        try:
            self.widget.mpl.canvas.draw_idle()
        except Exception:
            pass

    def _set_row_roi_spin_values_from_pixels(self, y0, y1):
        img = getattr(self.model.diff_img, "img", None)
        if img is None or np.ndim(img) < 2:
            return
        ymin = int(np.floor(min(float(y0), float(y1))))
        ymax = int(np.ceil(max(float(y0), float(y1))))
        ymax = max(ymin, ymax - 1)
        n_rows = int(img.shape[0])
        ymin = max(0, min(ymin, n_rows - 1))
        ymax = max(0, min(ymax, n_rows - 1))
        self.widget.spinBox_CCDRowMin.blockSignals(True)
        self.widget.spinBox_CCDRowMax.blockSignals(True)
        self.widget.spinBox_CCDRowMin.setValue(ymin)
        self.widget.spinBox_CCDRowMax.setValue(ymax)
        self.widget.spinBox_CCDRowMin.blockSignals(False)
        self.widget.spinBox_CCDRowMax.blockSignals(False)

    def _reset_row_roi_to_full(self):
        img = getattr(getattr(self.model, "diff_img", None), "img", None)
        if img is None or np.ndim(img) < 2 or img.shape[0] <= 1:
            return
        self.widget.spinBox_CCDRowMin.blockSignals(True)
        self.widget.spinBox_CCDRowMax.blockSignals(True)
        self.widget.spinBox_CCDRowMin.setValue(0)
        self.widget.spinBox_CCDRowMax.setValue(int(img.shape[0] - 1))
        self.widget.spinBox_CCDRowMin.blockSignals(False)
        self.widget.spinBox_CCDRowMax.blockSignals(False)
        self._apply_row_roi_to_spectrum()

    """
    def load_cake_format_file(self):
        # get filename
        temp_dir = get_temp_dir(self.model.get_base_ptn_filename())
        filen = QtWidgets.QFileDialog.getOpenFileName(
            self.widget, "Open a cake format File", temp_dir,  # self.model.chi_path,
            "Data files (*.cakeformat)")[0]
        if filen == '':
            return
        temp_values = []
        with open(filen, "r") as f:
            for line in f:
                temp_values.append(float(line.split(':')[1]))
        self.widget.spinBox_AziShift.setValue(temp_values[0])
        self.widget.spinBox_MaxCakeScale.setValue(temp_values[1])
        self.widget.horizontalSlider_VMin.setValue(temp_values[2])
        self.widget.horizontalSlider_VMax.setValue(temp_values[3])
        self.widget.horizontalSlider_MaxScaleBars.setValue(temp_values[4])
        self._apply_changes_to_graph()

    def save_cake_format_file(self):
        # make filename
        temp_dir = get_temp_dir(self.model.get_base_ptn_filename())
        ext = "cakeformat"
        #filen_t = self.model.make_filename(ext)
        filen_t = make_filename(self.model.base_ptn.fname, ext,
                                temp_dir=temp_dir)
        filen = dialog_savefile(self.widget, filen_t)
        if str(filen) == '':
            return
        # save cake related Values
        names = ['azi_shift', 'int_max', 'min_bar', 'max_bar', 'scale_bar']
        values = [self.widget.spinBox_AziShift.value(),
                  self.widget.spinBox_MaxCakeScale.value(),
                  self.widget.horizontalSlider_VMin.value(),
                  self.widget.horizontalSlider_VMax.value(),
                  self.widget.horizontalSlider_MaxScaleBars.value()]

        with open(filen, "w") as f:
            for n, v in zip(names, values):
                f.write(n + ' : ' + str(v) + '\n')
    """

    def reset_max_cake_scale(self):
        if hasattr(self.widget, "checkBox_Diff") and self.widget.checkBox_Diff.isChecked():
            return
        self._ensure_row_roi_defined_for_full_ccd()
        self._set_row_roi_spin_limits()
        intensity_cake, _, _ = self.model.diff_img.get_cake()
        arr = np.asarray(intensity_cake, dtype=float)
        arr = arr[np.isfinite(arr)]
        if arr.size == 0:
            return
        vmin = float(np.min(arr))
        vmax = float(np.max(arr))
        self._set_cake_scale_spinboxes(vmin, vmax)
        self._apply_changes_to_graph()

    def _apply_changes_to_graph(self):
        self.plot_ctrl.update()

    def _set_cake_bound_from_hist(self, bound_type, intensity_value):
        if bound_type == "min":
            self.widget.doubleSpinBox_CCDScaleMin.setValue(float(intensity_value))
        elif bound_type == "max":
            self.widget.doubleSpinBox_CCDScaleMax.setValue(float(intensity_value))

    def _set_cake_scale_spinboxes(self, vmin, vmax):
        if not hasattr(self.widget, "doubleSpinBox_CCDScaleMin"):
            return
        if vmax < vmin:
            vmin, vmax = vmax, vmin
        self.widget.doubleSpinBox_CCDScaleMin.blockSignals(True)
        self.widget.doubleSpinBox_CCDScaleMax.blockSignals(True)
        self.widget.doubleSpinBox_CCDScaleMin.setValue(float(vmin))
        self.widget.doubleSpinBox_CCDScaleMax.setValue(float(vmax))
        self.widget.doubleSpinBox_CCDScaleMin.blockSignals(False)
        self.widget.doubleSpinBox_CCDScaleMax.blockSignals(False)

    def _on_cake_scale_spin_changed(self, value):
        del value
        if not hasattr(self.widget, "doubleSpinBox_CCDScaleMin"):
            return
        vmin = float(self.widget.doubleSpinBox_CCDScaleMin.value())
        vmax = float(self.widget.doubleSpinBox_CCDScaleMax.value())
        if vmax <= vmin:
            vmax = vmin + max(1e-6, 1e-6 * max(abs(vmin), 1.0))
            self.widget.doubleSpinBox_CCDScaleMax.blockSignals(True)
            self.widget.doubleSpinBox_CCDScaleMax.setValue(vmax)
            self.widget.doubleSpinBox_CCDScaleMax.blockSignals(False)
        self._apply_changes_to_graph()

    def _ignore_raw_data_missing(self):
        return True

    def _set_image_file_box_missing(self):
        self.widget.textEdit_DiffractionImageFilename.setText(
            'CCD image file is missing. Move the image file into the same folder as the spectrum.')

    def _set_image_file_box(self):
        if self.model.diff_img_exist() and (self.model.diff_img.img_filename is not None):
            self.widget.textEdit_DiffractionImageFilename.setText(
                self.model.diff_img.img_filename)
        else:
            self._set_image_file_box_missing()

    def _warn_cannot_process_cake(self):
        QtWidgets.QMessageBox.warning(
            self.widget, 'Warning',
            'Rampo cannot process the CCD view: no image data and no cached CCD files were found.')

    def _load_cake_from_temp_without_raw_image(self, temp_dir):
        # Temp cake files are named from the base pattern root name.
        if self.model.diff_img is None:
            self.model.reset_diff_img()
        self.model.diff_img.img_filename = self.model.make_filename(
            'tif', original=True)
        return self.model.diff_img.read_cake_from_tempfile(temp_dir=temp_dir)

    def show_tif_header(self):
        if not self.model.base_ptn_exist():
            return
        filen_tif = self.model.make_filename('tif', original=True)
        filen_tiff = self.model.make_filename('tiff', original=True)
        if not (os.path.exists(filen_tif) or
                os.path.exists(filen_tiff)):
            QtWidgets.QMessageBox.warning(
                self.widget, 'Warning',
                'Cannot find image file: %s or %s in the chi folder.' %
                (filen_tif, filen_tiff))
        else:
            textoutput = ''
            if os.path.exists(filen_tif):
                f = filen_tif
            else:
                f = filen_tiff
            metadata = {}
            with Image.open(f) as img:
                for key in img.tag:
                    metadata[key] = img.tag[key]
            infobox = InformationBox()
            infobox.setText(json.dumps(metadata, indent=4))
            print(str(datetime.datetime.now())[:-7], ': TIF metadata\n', 
                json.dumps(metadata, indent=4))
            infobox.exec()
            #self.widget.plainTextEdit_ViewJCPDS.setPlainText(textoutput)


    def addremove_cake(self):
        """
        add / remove cake to the graph
        """
        update = self._addremove_cake()
        if update:
            self._apply_changes_to_graph()
    """
    def image_file_exists(self):
        # if no image file, no cake
        filen_tif = self.model.make_filename('tif', original=True)
        filen_tiff = self.model.make_filename('tiff', original=True)
        filen_mar3450 = self.model.make_filename('mar3450',
            original=True)
        filen_cbf = self.model.make_filename('cbf', original=True)
        filen_h5 = self.model.make_filename('h5', original=True)
        if not (os.path.exists(filen_tif) or
                os.path.exists(filen_tiff) or
                os.path.exists(filen_mar3450) or
                os.path.exists(filen_h5) or
                os.path.exists(filen_cbf)):
            return False
        else:
            return True
    """

    def _addremove_cake(self):
        """
        add / remove cake
        no signal to update_graph
        """
        # if no base ptn, no cake
        if not self.model.base_ptn_exist():
            QtWidgets.QMessageBox.warning(
                self.widget, 'Warning', 'Choose a spectrum file first.')
            return False
        if self._is_spe_source():
            if not self.process_temp_cake():
                return False
            return True
        if not self.model.associated_image_exists():
            if not self._ignore_raw_data_missing():
                self._set_image_file_box_missing()
                QtWidgets.QMessageBox.warning(
                    self.widget, 'Warning',
                    'Cannot find CCD image file.')
                return False
            if not self.process_temp_cake():
                self._warn_cannot_process_cake()
                return False
            return True

        if not self.process_temp_cake():
            self._warn_cannot_process_cake()
            return False
        return True

    def _load_new_image(self):
        """
        Load new image for cake view.  Cake should be the same as base pattern.
        no signal to update_graph
        """
        self.model.reset_diff_img()
        if not self.model.associated_image_exists():
            self._set_image_file_box_missing()
            return False
        self.model.load_associated_img()
        self._set_image_file_box()
        return True

    def _is_spe_source(self):
        if not self.model.base_ptn_exist():
            return False
        return extract_extension(str(self.model.get_base_ptn_filename())).lower() == 'spe'

    def apply_mask(self):
        # self.produce_cake()
        min_mask = float(self.widget.spinBox_MaskMin.value())
        max_mask = float(self.widget.spinBox_MaskMax.value())
        zrange = self.model.diff_img.get_img_zrange()
        print('img z range', zrange)
        print('mask range', min_mask, max_mask)
        if (zrange[0] < min_mask) or (zrange[1] > max_mask):
            # case for meaningful mask
            if self.widget.pushButton_ApplyMask.isChecked():
                self.ccdprocess_ctrl.cook()
        else:
            self.model.diff_img.set_mask(None)
        self._apply_changes_to_graph()

    def reset_maskrange(self):
        # get min and max of the cake image
        #intensity_cake, _, _ = self.model.diff_img.get_cake()
        zrange = self.model.diff_img.get_img_zrange()
        if zrange != None:
            # push those values to spinboxes
            self.widget.spinBox_MaskMin.setValue(int(zrange[0]))
            self.widget.spinBox_MaskMax.setValue(int(zrange[1]))
            self.model.diff_img.set_mask(None)
            self._apply_changes_to_graph()
        # reprocess the image
        # self.apply_mask()

    def produce_cake(self):
        """
        Reprocess to get cake.  Slower re - processing
        does not signal to update_graph
        """
        success = self._load_new_image()
        if not success:
            return False
        self._set_row_roi_spin_limits()
        self.ccdprocess_ctrl.cook()
        self._apply_row_roi_to_spectrum()
        self._set_image_file_box()
        return True

    def process_temp_cake(self):
        """
        load cake through either temporary file or make a new cake
        """
        temp_dir = get_temp_dir(self.model.get_base_ptn_filename())
        if self._is_spe_source():
            success = self._load_new_image()
            if not success:
                return False
            self._set_row_roi_spin_limits()
            self._apply_row_roi_to_spectrum()
            self.model.diff_img.write_temp_cakefiles(temp_dir=temp_dir)
            return True
        has_raw_image = self.model.associated_image_exists()
        if not has_raw_image:
            self._set_image_file_box_missing()
            if not self._ignore_raw_data_missing():
                QtWidgets.QMessageBox.warning(
                    self.widget, "Warning",
                    "Image file for the base pattern does not exist.")
                return False
            return self._load_cake_from_temp_without_raw_image(temp_dir)
        #temp_dir = os.path.join(self.model.chi_path, 'temporary_pkpo')
        success = self._load_new_image()
        if not success:
            return False
        success = self.model.diff_img.read_cake_from_tempfile(
            temp_dir=temp_dir)
        if success:
            print(str(datetime.datetime.now())[:-7],
                ": Load cake image from temporary file.")
            return True
        print(str(datetime.datetime.now())[:-7],
            ": Create new temporary file for cake image.")
        self._update_temp_cake_files(temp_dir)
        return True

    def _update_temp_cake_files(self, temp_dir):
        success = self.produce_cake()
        if not success:
            return
        self.model.diff_img.write_temp_cakefiles(temp_dir=temp_dir)

    def temp_cake_exists(self):
        temp_dir = get_temp_dir(self.model.get_base_ptn_filename())
        search_pattern = os.path.join(temp_dir, "*.cake.npy")
        cake_all = glob.glob(search_pattern)
        print(len(cake_all))
        if len(cake_all) < 3:
            return False
        else:
            return True


CakeController = CCDController



import os
import traceback
import numpy as np
from qtpy import QtWidgets
from .mplcontroller import MplController
from .peakfittablecontroller import PeakfitTableController
from ..utils import make_filename, get_temp_dir
from ..model.param_session_io import load_section_from_param


class PeakFitController(object):

    def __init__(self, model, widget):
        self.model = model
        self.widget = widget
        self.plot_ctrl = MplController(self.model, self.widget)
        self.peakfit_table_ctrl = PeakfitTableController(
            self.model, self.widget)
        self.connect_channel()

    def connect_channel(self):
        self.widget.pushButton_SetFitSection.clicked.connect(
            self.set_fit_section)
        self.widget.pushButton_ConductFitting.clicked.connect(
            self.conduct_fitting)
        self.widget.pushButton_PkSave.clicked.connect(
            self.save_to_section)
        self.widget.pushButton_ClearSection.clicked.connect(
            self.clear_this_section)
        self.widget.pushButton_PkFtSectionRemove.clicked.connect(
            self.remove_section)
        self.widget.pushButton_PkFtSectionSetToCurrent.clicked.\
            connect(self.set_section_to_current)
        self.widget.pushButton_ZoomToSection.clicked.connect(
            self.zoom_to_section)
        self.widget.pushButton_PkFtSectionSavetoXLS.clicked.\
            connect(self.save_to_xls)
        self.widget.pushButton_PkFtSectionImport.clicked.connect(
            self.import_section_from_param)
        # The line below exist in session_ctrl
        # self.widget.pushButton_PkFtSectionSavetoDPP.clicked.coonect

    def import_section_from_param(self):
        if not self.model.base_ptn_exist():
            QtWidgets.QMessageBox.warning(
                self.widget, "Warning", "Open a base spectrum first.")
            return
        idx, ok = QtWidgets.QInputDialog.getInt(
            self.widget,
            "Import Section",
            "Section index:",
            0,
            0,
            max(0, len(self.model.section_lst) + 1000),
            1,
        )
        if not ok:
            return
        try:
            section = load_section_from_param(
                self.model.get_base_ptn_filename(), int(idx))
        except Exception as inst:
            err = traceback.format_exc()
            print(err)
            QtWidgets.QMessageBox.warning(
                self.widget, "Warning", str(inst) + "\n\n" + err)
            return
        if section is None:
            QtWidgets.QMessageBox.warning(
                self.widget, "Warning",
                "No saved section was found for that index in the current session folder.")
            return
        self.model.section_lst.append(section)
        self.peakfit_table_ctrl.update_sections()

    def import_section_from_dpp(self):
        return self.import_section_from_param()

    def zoom_to_section(self):
        if not self.model.current_section_exist():
            return
        x_range = self.model.current_section.get_xrange()
        y_range = self.model.current_section.get_yrange(
            bgsub=self.widget.checkBox_BgSub.isChecked())
        margin = 0.1 * (y_range[1] - y_range[0])
        self.plot_ctrl.update(
            limits=(x_range[0], x_range[1],
                    y_range[0] - margin, y_range[1] + margin))

    def get_style_for_unsaved(self):
        return "Background-color:rgb(255,204,255);color:rgb(0,0,0);"

    def get_style_for_saved(self):
        return "Background-color:None;color:rgb(0,0,0);"

    def set_tableWidget_PkParams_saved(self):
        self.widget.tableWidget_PkParams.setStyleSheet(
            self.get_style_for_saved())

    def set_tableWidget_PkParams_unsaved(self):
        self.widget.tableWidget_PkParams.setStyleSheet(
            self.get_style_for_unsaved())

    def set_tableWidget_PkFtSections_saved(self):
        self.widget.tableWidget_PkFtSections.setStyleSheet(
            self.get_style_for_saved())

    def set_tableWidget_PkFtSections_unsaved(self):
        self.widget.tableWidget_PkFtSections.setStyleSheet(
            self.get_style_for_unsaved())

    def set_section_to_current(self):
        if self.widget.tableWidget_PkFtSections.selectionModel().\
                selectedRows().__len__() != 1:
            QtWidgets.QMessageBox.warning(
                self.widget, 'Warning',
                'Select a row to make current.')
            return
        if self.model.current_section_exist():
            if not self.model.current_section_saved():
                reply = QtWidgets.QMessageBox.question(
                    self.widget, 'Message',
                    'Are you OK to loose current section information?',
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                    QtWidgets.QMessageBox.Yes)
                if reply == QtWidgets.QMessageBox.No:
                    return
                else:
                    pass
            else:
                pass
        else:
            pass
        idx = self.widget.tableWidget_PkFtSections.selectionModel().\
            selectedRows()[0].row()
        # Reload selected section from PARAM CSV on demand so graph updates
        # reflect persisted section data (not stale in-memory copies).
        if self.model.base_ptn_exist():
            try:
                section_disk = load_section_from_param(
                    self.model.get_base_ptn_filename(), idx)
            except Exception:
                section_disk = None
            if section_disk is not None:
                self.model.section_lst[idx] = section_disk
            else:
                QtWidgets.QMessageBox.warning(
                    self.widget, "Missing Section CSV",
                    "Saved section CSV for the selected row was not found.\n"
                    "Using in-memory section data instead."
                )
        self.model.set_this_section_current(idx)
        self.peakfit_table_ctrl.update_peak_parameters()
        self.peakfit_table_ctrl.update_sections()
        self.peakfit_table_ctrl.update_peak_constraints()
        """
        self._list_peaks()
        self._list_localbg()
        self._update_config
        """
        self.zoom_to_section()

    def clear_section_list(self):
        reply = QtWidgets.QMessageBox.question(
            self.widget, 'Message',
            'Any unsaved sections will be erased. Proceed?',
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.Yes)
        if reply == QtWidgets.QMessageBox.No:
            return
        self._clear_all_sections()

    def _clear_all_sections(self):
        '''clean both sections and currentSection'''
        self.model.clear_section_list()
        self.widget.tableWidget_PkFtSections.clearContents()
        self._clear_current_section()

    def remove_section(self):
        reply = QtWidgets.QMessageBox.question(
            self.widget, 'Message',
            'Are you sure you want to delete the selected sections?',
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.Yes)
        if reply == QtWidgets.QMessageBox.No:
            return
        idx_checked = []
        for item in \
                self.widget.tableWidget_PkFtSections.selectionModel().\
                selectedRows():
            idx_checked.append(item.row())
        # remove checked ones
        if idx_checked != []:
            idx_checked.reverse()
            for idx in idx_checked:
                self.model.section_lst.pop(idx)
                self.widget.tableWidget_PkFtSections.removeRow(idx)

    def clear_this_section(self):
        reply = QtWidgets.QMessageBox.question(
            self.widget, 'Message',
            'Any unsaved fitting result will be discarded.  Proceed?',
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.Yes)
        limits = self.widget.mpl.canvas.ax_pattern.axis()
        if reply == QtWidgets.QMessageBox.No:
            return
        else:
            self._clear_current_section()
            # self.plot_ctrl.update(limits, False)
            self.plot_ctrl.update()

    def _clear_current_section(self):
        '''erase only current section'''
        self.widget.tableWidget_PeakConstraints.clearContents()
        self.widget.tableWidget_PkParams.clearContents()
        self.model.initialize_current_section()

    def save_to_section(self):
        if not self.model.current_section_exist():
            return
        self.model.save_current_section()
        self.set_tableWidget_PkParams_saved()
        self.set_tableWidget_PkFtSections_unsaved()
        # self._list_sections()
        self.model.initialize_current_section()
        self.widget.tableWidget_PkParams.clearContents()
        self.widget.tableWidget_PeakConstraints.clearContents()
        self.peakfit_table_ctrl.update_sections()
        self.plot_ctrl.update()

    def set_fit_section(self):
        # if there is unsaved section, ask if it needs to be saved,
        # this can be checked by looking at timestamp
        if self.model.current_section_exist():
            if not self.model.current_section_saved():
                reply = QtWidgets.QMessageBox.question(
                    self.widget, 'Message',
                    'Do you want to save the unsaved section?',
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                    QtWidgets.QMessageBox.Yes)
                if reply == QtWidgets.QMessageBox.Yes:
                    self.model.save_current_section()
        self.model.initialize_current_section()
        # Warn the users about what is going to happen
        reply = QtWidgets.QMessageBox.question(
            self.widget, 'Message',
            'I will set current plot range as current section. OK?',
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.Yes)
        if reply == QtWidgets.QMessageBox.No:
            return
        # get X range from the current view
        axisrange = self.widget.mpl.canvas.ax_pattern.axis()
        self.model.set_current_section([axisrange[0], axisrange[1]])
        # This button itself should activate the mouse for peak picking model
        self.widget.pushButton_AddRemoveFromMouse.setChecked(True)
        self.set_tableWidget_PkParams_unsaved()

    def set_mouse_for_peak_input(self):
        if self.widget.pushButton_AddRemoveFromMouse.isChecked():
            self.release_mouse_from_peak_input()
        else:
            self.widget.pushButton_AddRemoveFromMouse.setChecked(True)

    def release_mouse_from_peak_input(self):
        if self.widget.mpl.ntb._active == 'PAN':
            self.widget.mpl.ntb.pan()
        if self.widget.mpl.ntb._active == 'ZOOM':
            self.widget.mpl.ntb.zoom()
        self.widget.pushButton_AddRemoveFromMouse.setChecked(False)
        return
    '''
    def pick_peak(self, mouse_button, xdata, ydata):
        """
        """
        if mouse_button == 'left':  # left click
            success = self.model.current_section.set_single_peak(
                float(xdata),
                self.widget.doubleSpinBox_InitialFWHM.value())
            if not success:
                QtWidgets.QMessageBox.warning(
                    self.widget, "Warning",
                    "You picked outside of the current section.")
                return
        elif mouse_button == 'right':  # right button for removal
            if not self.model.current_section.peaks_exist():
                return
            self.model.current_section.remove_single_peak_nearby(xdata)
        else:
            return
        self.set_tableWidget_PkParams_unsaved()
        self.peakfit_table_ctrl.update_peak_parameters()
        self.peakfit_table_ctrl.update_peak_constraints()
        self.plot_ctrl.update()
    '''

    def conduct_fitting(self):
        if not self.model.current_section_exist():
            QtWidgets.QMessageBox.warning(
                self.widget, "Warning", "No section is defined")
            return
        if self.model.current_section.get_number_of_peaks_in_queue() == 0:
            QtWidgets.QMessageBox.warning(
                self.widget, "Warning", "No pick exists in the section.")
            return
        width = self.widget.doubleSpinBox_InitialFWHM.value()
        order = self.widget.spinBox_BGPolyOrder.value()
        maxwidth = self.widget.doubleSpinBox_MaxFWHM.value()
        centerrange = self.widget.doubleSpinBox_PeakCenterRange.value()
        self.model.current_section.prepare_for_fitting(order,
                                                       maxwidth, centerrange)
        success = self.model.current_section.conduct_fitting()
        if success:
            QtWidgets.QMessageBox.warning(self.widget, "Information",
                                          'Fitting finished.')
            self.plot_ctrl.update(limits=self._fit_result_limits())
            self.peakfit_table_ctrl.update_peak_parameters()
            self.peakfit_table_ctrl.update_peak_constraints()
            self.set_tableWidget_PkParams_unsaved()
        else:
            QtWidgets.QMessageBox.warning(self.widget, "Information",
                                          'Fitting failed.')

    def _fit_result_limits(self):
        if not self.model.current_section_exist():
            return None
        section = self.model.current_section
        bgsub = self.widget.checkBox_BgSub.isChecked()
        x_range = section.get_xrange()
        try:
            current_limits = self.widget.mpl.canvas.ax_pattern.axis()
            x_min = max(float(current_limits[0]), float(x_range[0]))
            x_max = min(float(current_limits[1]), float(x_range[1]))
            if x_max <= x_min:
                x_min, x_max = float(x_range[0]), float(x_range[1])
        except Exception:
            x_min, x_max = float(x_range[0]), float(x_range[1])

        y_min, y_max = section.get_yrange(bgsub=bgsub)
        residue = section.get_fit_residue(bgsub=bgsub)
        y_shift = y_min - (y_max - y_min) * 0.05
        residue_plot = np.asarray(residue, dtype=float) + y_shift
        if residue_plot.size > 0:
            y_min = min(float(y_min), float(np.min(residue_plot)))
            y_max = max(float(y_max), float(np.max(residue_plot)))
        if y_max <= y_min:
            span = max(abs(y_max), 1.0) * 1.0e-6
        else:
            span = y_max - y_min
        margin = span * 0.05
        return (x_min, x_max, y_min - margin, y_max + margin)

    def save_to_xls(self):
        temp_dir = get_temp_dir(self.model.get_base_ptn_filename())
        filen_xls = make_filename(self.model.get_base_ptn_filename(),
                      'peakfit.xls', temp_dir=temp_dir)
        reply = QtWidgets.QMessageBox.question(
            self.widget, 'Question',
            'Do you want to save in default filename, %s ?' % filen_xls,
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.Yes)
        if reply == QtWidgets.QMessageBox.No:
            filen_xls = QtWidgets.QFileDialog.getSaveFileName(
                self.widget, "Save an Excel File", filen_xls, "(*.xls)")
        else:
            if os.path.exists(filen_xls):
                reply = QtWidgets.QMessageBox.question(
                    self.widget, 'Question',
                    'The file already exist.  Overwrite?',
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                    QtWidgets.QMessageBox.No)
                if reply == QtWidgets.QMessageBox.No:
                    return
        self.model.save_peak_fit_results_to_xls(filen_xls)



import os
from qtpy import QtWidgets
from ..utils import get_sorted_filelist, find_from_filelist, readchi, \
    make_filename, writechi, get_directory
from ..utils import undo_button_press, get_temp_dir
import datetime
from .mplcontroller import MplController
from .ccdcontroller import CCDController


class BaseSpectrumController(object):

    def __init__(self, model, widget, session_ctrl=None):
        self.model = model
        self.widget = widget
        self.session_ctrl = session_ctrl
        self.plot_ctrl = MplController(self.model, self.widget)
        self.ccd_ctrl = CCDController(self.model, self.widget)
        self.connect_channel()

    def connect_channel(self):
        self.widget.pushButton_NewBasePtn.clicked.connect(
            self.select_base_ptn)
        self.widget.lineEdit_DiffractionPatternFileName.editingFinished.\
            connect(self.load_new_base_pattern_from_name)

    def select_base_ptn(self):
        """
        opens a file select dialog
        """
        filen = QtWidgets.QFileDialog.getOpenFileName(
            self.widget, "Open an SPE or CHI File", self.model.chi_path,
            "Data files (*.spe *.SPE *.chi)")[0]
        self._setshow_new_base_ptn(str(filen))

    def load_new_base_pattern_from_name(self):
        if self.widget.lineEdit_DiffractionPatternFileName.isModified():
            filen = self.widget.lineEdit_DiffractionPatternFileName.text()
            self._setshow_new_base_ptn(filen)

    def _setshow_new_base_ptn(self, filen):
        """
        load and then send signal to update_graph
        """
        if os.path.exists(filen):
            self.model.set_chi_path(os.path.split(filen)[0])
            if self.model.base_ptn_exist():
                old_filename = self.model.get_base_ptn_filename()
            else:
                old_filename = None
            new_filename = filen
            self._load_a_new_pattern(new_filename)
            if old_filename is None:
                self.plot_new_graph()
            else:
                self.apply_changes_to_graph()
        else:
            QtWidgets.QMessageBox.warning(
                self.widget, 'Warning', 'Cannot find ' + filen)
            # self.widget.lineEdit_DiffractionPatternFileName.setText(
            #    self.model.get_base_ptn_filename())

    def _load_a_new_pattern(self, new_filename):
        """
        load and process base pattern.  does not signal to update_graph
        """
        if self.session_ctrl is not None:
            # Reset carry-over provenance for generic/manual loads.
            self.session_ctrl.set_carryover_source_chi(None)
        self.model.set_base_ptn(
            new_filename, self.widget.doubleSpinBox_SetWavelength.value())
        # self.widget.textEdit_DiffractionPatternFileName.setText(
        #    '1D Pattern: ' + self.model.get_base_ptn_filename())
        self.widget.lineEdit_DiffractionPatternFileName.setText(
            str(self.model.get_base_ptn_filename()))
        main_ctrl = getattr(self.widget, "_main_controller", None)
        if main_ctrl is not None and hasattr(main_ctrl, "sync_background_roi_spinboxes"):
            try:
                main_ctrl.sync_background_roi_spinboxes(force_full_range=True)
            except Exception:
                pass
        # Prefer loading full PARAM session state when available for this CHI.
        if self.session_ctrl is not None:
            loaded_param = self.session_ctrl.autoload_param_for_chi(new_filename)
            if loaded_param:
                if self.ccd_ctrl._is_spe_source():
                    try:
                        self.ccd_ctrl.process_temp_cake()
                    except Exception:
                        pass
                # Ensure File > Data backup table reflects the newly loaded CHI
                # immediately, without requiring tab changes.
                self.session_ctrl.refresh_backup_table()
                print(str(datetime.datetime.now())[:-7],
                    ': Loaded PARAM session for this CHI.')
                return
        print(str(datetime.datetime.now())[:-7], 
                ": Receive request to open ", 
                str(self.model.get_base_ptn_filename()))
        temp_dir = get_temp_dir(self.model.get_base_ptn_filename())
        if True:
            if os.path.exists(temp_dir):
                success = self.model.base_ptn.read_bg_from_tempfile(
                    temp_dir=temp_dir)
                if success:
                    self._update_bg_params_in_widget()
                    print(str(datetime.datetime.now())[:-7], 
                        ': Read temp chi successfully.')
                else:
                    self._update_bgsub_from_current_values()
                    print(str(datetime.datetime.now())[:-7], 
                        ': No temp chi file found. Force new bgsub fit.')
            else:
                os.makedirs(temp_dir)
                self._update_bgsub_from_current_values()
                print(str(datetime.datetime.now())[:-7], 
                    ': No temp chi file found. Force new bgsub fit.')
        if (not self.ccd_ctrl._is_spe_source()) and \
                (not self.model.associated_image_exists()) and \
                (not self.ccd_ctrl._ignore_raw_data_missing()):
            return

        success = self.ccd_ctrl.process_temp_cake()
        if (not success) and \
                self.ccd_ctrl._ignore_raw_data_missing() and \
                (not self.model.associated_image_exists()):
            QtWidgets.QMessageBox.warning(
                self.widget, 'Warning',
                'Rampo cannot process the CCD view: no raw image and no existing cached files were found.')
        # Keep backup table in File > Data synchronized right after CHI load.
        if self.session_ctrl is not None:
            self.session_ctrl.refresh_backup_table()

    def _update_bg_params_in_widget(self):
        self.widget.spinBox_BGParam1.setValue(
            self.model.base_ptn.params_chbg[0])
        self.widget.doubleSpinBox_Background_ROI_min.setValue(
            self.model.base_ptn.roi[0])
        self.widget.doubleSpinBox_Background_ROI_max.setValue(
            self.model.base_ptn.roi[1])
        table = getattr(self.widget, "tableWidget_BackgroundConstraints", None)
        if table is not None:
            table.setRowCount(0)
            for area in getattr(self.model.base_ptn, "bg_fit_areas", []) or []:
                try:
                    xmin = float(area[0])
                    xmax = float(area[1])
                except Exception:
                    continue
                row = table.rowCount()
                table.insertRow(row)
                table.setItem(row, 0, QtWidgets.QTableWidgetItem(f"{xmin:.3f}"))
                table.setItem(row, 1, QtWidgets.QTableWidgetItem(f"{xmax:.3f}"))

    def _update_bgsub_from_current_values(self):
        fit_areas = []
        table = getattr(self.widget, "tableWidget_BackgroundConstraints", None)
        if table is not None:
            for row in range(table.rowCount()):
                item_min = table.item(row, 0)
                item_max = table.item(row, 1)
                if item_min is None or item_max is None:
                    continue
                try:
                    xmin = float(item_min.text())
                    xmax = float(item_max.text())
                except Exception:
                    continue
                if xmax < xmin:
                    xmin, xmax = xmax, xmin
                fit_areas.append([xmin, xmax])
        x_raw, y_raw = self.model.base_ptn.get_raw()
        __, y_fit = self.plot_ctrl._get_smoothed_pattern_xy(x_raw, y_raw)
        if not self.plot_ctrl._smoothing_active():
            y_fit = y_raw
        if (x_raw.min() >= self.widget.doubleSpinBox_Background_ROI_min.value()) or \
                (x_raw.max() <= self.widget.doubleSpinBox_Background_ROI_min.value()):
            self.widget.doubleSpinBox_Background_ROI_min.setValue(x_raw.min())
        if (x_raw.max() <= self.widget.doubleSpinBox_Background_ROI_max.value()) or \
                (x_raw.min() >= self.widget.doubleSpinBox_Background_ROI_max.value()):
            self.widget.doubleSpinBox_Background_ROI_max.setValue(x_raw.max())
        self.model.base_ptn.subtract_bg(
            [self.widget.doubleSpinBox_Background_ROI_min.value(),
                self.widget.doubleSpinBox_Background_ROI_max.value()],
            [self.widget.spinBox_BGParam1.value()],
            yshift=0,
            fit_areas=fit_areas,
            y_source=y_fit)
        temp_dir = get_temp_dir(self.model.get_base_ptn_filename())
        self.model.base_ptn.write_temporary_bgfiles(temp_dir)

    def apply_changes_to_graph(self):
        self.plot_ctrl.update()

    def plot_new_graph(self):
        self.plot_ctrl.zoom_out_graph()


BasePatternController = BaseSpectrumController



from qtpy import QtWidgets


def undo_button_press(buttonObj, released_text='Released',
                      pressed_text='Pressed'):
    """
    recover checkable pushButton status when error occurs

    :param buttonObj: pyqt4 pushButton object
    :param released_text: text label of the button when released
    :param pressed_text: text label of the button when pressed
    """
    if buttonObj.isChecked():
        buttonObj.setChecked(False)
        buttonObj.setText(released_text)
    else:
        buttonObj.setChecked(True)
        buttonObj.setText(pressed_text)
    return


class SpinBoxFixStyle(QtWidgets.QProxyStyle):
    """
    Copied from https://stackoverflow.com/questions/40746350/why-qspinbox-jumps-twice-the-step-value
    To fix the SpinBox button problem.  This fixes SpinBoxes issuing events
    twice.
    """

    def styleHint(self, hint, option=None, widget=None, returnData=None):
        if hint == QtWidgets.QStyle.SH_SpinBox_KeyPressAutoRepeatRate:
            return 5**10
        elif hint == QtWidgets.QStyle.SH_SpinBox_ClickAutoRepeatRate:
            return 5**10
        elif hint == QtWidgets.QStyle.SH_SpinBox_ClickAutoRepeatThreshold:
            # You can use only this condition to avoid the auto-repeat,
            # but better safe than sorry ;-)
            return 5**10
        else:
            return super().styleHint(hint, option, widget, returnData)
